// This file is a part of the OpenSurgSim project.
// Copyright 2013, SimQuest Solutions Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#ifndef SURGSIM_MATH_SHAPES_H
#define SURGSIM_MATH_SHAPES_H

/// This file includes all the shapes
#include "SurgSim/Math/BoxShape.h"
#include "SurgSim/Math/CapsuleShape.h"
#include "SurgSim/Math/CylinderShape.h"
#include "SurgSim/Math/DoubleSidedPlaneShape.h"
#include "SurgSim/Math/MeshShape.h"
#include "SurgSim/Math/PlaneShape.h"
#include "SurgSim/Math/OctreeShape.h"
#include "SurgSim/Math/SphereShape.h"
#include "SurgSim/Math/SurfaceMeshShape.h"

#endif // SURGSIM_MATH_SHAPES_H
